from collections.abc import Iterable
from typing import Optional

import astropy.units as u
import numpy as np
from astropy.time import Time
import re


def flatten(lst):
    """
    Flattens any iterable, no matter how irregular.
    Example: flatten([1, 2, [3, 4, 5], [[6, 7], 8, [9, 10]]])
    Deliberately left untyped to allow for maximum type usage.
    """
    for el in lst:
        if isinstance(el, Iterable) and not isinstance(el, (str, bytes)):
            yield from flatten(el)
        else:
            yield el


def round_minute(time: Time, up: bool = False) -> Time:
    """
    Round a time down (truncate) or up to the nearest minute
    time: an astropy.Time
    up: bool indicating whether to round up
    """
    t = time.copy()
    t.format = 'iso'
    t.out_subfmt = 'date_hm'
    if up:
        sec = t.strftime('%S').astype(int)
        idx = np.where(sec > 0)
        t[idx] += 1.0 * u.min
    return Time(t.iso, format='iso', scale='utc')


def str_to_bool(s: Optional[str]) -> bool:
    """
    Returns true if and only if s is defined and some variant capitalization of 'yes' or 'true'.
    """
    return s is not None and s.strip().upper() in ['YES', 'TRUE']


# A dict of signs for conversion.
SIGNS = {'': 1, '+': 1, '-': -1}


def dmsstr2deg(s: str) -> float:
    if not s:
        raise ValueError(f'Illegal DMS string: {s}')

    sign = '+'
    if s[0] in SIGNS:
        sign = s[0]
        s = s[1:]

    result = s.split(':')
    if len(result) != 3:
        raise ValueError(f'Illegal DMS string: {s}')

    try:
        return dms2deg(int(result[0]), int(result[1]), float(result[2]), sign)
    except ValueError:
        import logging
        logging.error(f'Uhoh: {s}')
        raise


def dms2deg(d: int, m: int, s: float, sign: str) -> float:
    if sign not in SIGNS:
        raise ValueError(f'Illegal sign "{sign}" in DMS: {sign}{d}:{m}:{s}')
    dec = SIGNS[sign] * (d + m / 60.0 + s / 3600.0)
    return dec if dec < 180 else -(360 - dec)


def dms2rad(d: int, m: int, s: float, sign: str) -> float:
    return dms2deg(d, m, s, sign) * np.pi / 180.0


def hmsstr2deg(s: str) -> float:
    if not s:
        raise ValueError(f'Illegal HMS string: {s}')

    result = s.split(':')
    if len(result) != 3:
        raise ValueError(f'Illegal HMS string: {s}')

    return hms2deg(int(result[0]), int(result[1]), float(result[2]))


def hms2deg(h: int, m: int, s: float) -> float:
    return h + m / 60.0 + s / 3600.0


def hms2rad(h: int, m: int, s: float) -> float:
    return hms2deg(h, m, s) * np.pi / 12.


def angular_distance(ra1: float, dec1: float, ra2: float, dec2: float) -> float:
    """
    Calculate the angular distance between two points on the sky.
    based on
    https://github.com/gemini-hlsw/lucuma-core/blob/master/modules/core/shared/src/main/scala/lucuma/core/math/Coordinates.scala#L52
    """
    phi_1 = dec1
    phi_2 = dec2
    delta_phi = dec2 - dec1
    delta_lambda = ra2 - ra1
    a = np.sin(delta_phi / 2)**2 + np.cos(phi_1) * np.cos(phi_2) * np.sin(delta_lambda / 2)**2
    return 2 * np.arctan2(np.sqrt(a), np.sqrt(1 - a))


barcodes = {'0.25arcsec': '10005371',
            '0.5arcsec': '10005372',
            '0.75arcsec': '10005373',
            '1.0arcsec': '10005374',
            '1.5arcsec': '10005375',
            '2.0arcsec': '10005376',
            '5.0arcsec': '10005377',
            'IFU-2': '10000007',
            'IFU-B': '10000008',
            'IFU-R': '10000009',
            'IFU-NS-2': '10000010',
            'IFU-NS-B': '10000011',
            'IFU-NS-R': '10000012',
            'NS0.5arcsec': '10005388',
            'NS0.75arcsec': '10005389',
            'NS1.0arcsec': '10005390',
            'NS1.5arcsec': '10005391',
            'NS2.0arcsec': '10005392',
            'PinholeC': '10005381',
            'focus_array_new': '10000005'}

inst_decode = {'GMOS': '1', 'F2': '3'}
sem_decode = {'A': '0', 'B': '1'}
prog_decode = {'Q': '0', 'C': '1', 'L': '2', 'F': '3', 'S': '8', 'D': '9'}


decoder = {'A': '0', 'B': '1', 'Q': '0',
           'C': '1', 'LP': '2', 'FT': '3',
           'SV': '8', 'DD': '9'}
bar_decoder = {value: key for key, value in decoder.items()}
pattern = '|'.join(decoder.keys())
sem_code = {value: key for key, value in sem_decode.items()}


def mask_to_barcode(mask: str, inst: Optional[str]) -> str:
    """
    Convert a mask string to a barcode string.
    """
    if mask in barcodes:
        return barcodes[mask]
    else:
        # return inst_decode[inst] + sem_decode[mask[6]] + prog_decode[mask[7]] + mask[-6:-3] + mask[-2:]
        return inst_decode[inst] + re.sub(f'({pattern})', lambda m: decoder[m.group()], mask).replace('-', '')[6:]


def barcode_to_mask(barcode: str, rootname: Optional[str]) -> str:
    """
    Convert a barcode string to a mask string.
    """
    
    if barcode in barcodes.values():
        return list(barcodes.keys())[list(barcodes.values()).index(barcode)]
    else:
        # Root is like 'GS2022'
        return rootname + sem_code[barcode[1]] + bar_decoder[barcode[2]] + barcode[3:6] + '-' + barcode[6:]



from datetime import timedelta

from ...minimodel import Observation

from .geminiproperties import GeminiProperties


def with_igrins_cal(func):
    def add_calibration(self):
        if GeminiProperties.Instruments.IGRINS in self.required_resources() and self.partner_used() > 0:
            return func(self) + timedelta(seconds=(1 / 6))
        return func(self)
    return add_calibration


class GeminiObservation(Observation):
    """
    A Gemini-specific extension of the Observation class.
    """
    @with_igrins_cal
    def total_used(self) -> timedelta:
        return super().total_used()



from typing import Tuple

import astropy.units as u
import numpy as np
from astropy.coordinates import Angle, SkyCoord, EarthLocation
from astropy.time import Time, TimeDelta

from .altitude import Altitude
from .constants import J2000
from .utils import current_geocent_frame, local_sidereal_time, hour_angle_to_angle


class Sun:
    @staticmethod
    def at(time: Time) -> SkyCoord:
        """
        low-precision position of the sun.

        Good to about 0.01 degree, from the 1990 Astronomical Almanac p. C24.
        At this level topocentric correction is not needed.

        Paramters
        ---------
        time : astropy Time

        Returns

        a SkyCoord in the geocentric frame of epoch of date.

        """

        # Low precision formulae for the sun, from Almanac p. C24 (1990) */
        # said to be good to about a hundredth of a degree.

        jd = np.asarray(time.jd)
        scalar_input = False
        if jd.ndim == 0:
            jd = jd[None]
            scalar_input = True

        n = jd - J2000  # referred to J2000
        ell = 280.460 + 0.9856474 * n
        g = np.deg2rad(357.528 + 0.9856003 * n)
        lambd = np.deg2rad(ell + 1.915 * np.sin(g) + 0.020 * np.sin(2. * g))
        epsilon = np.deg2rad(23.439 - 0.0000004 * n)

        x = np.cos(lambd)
        y = np.cos(epsilon) * np.sin(lambd)
        z = np.sin(epsilon) * np.sin(lambd)

        ra = np.arctan2(y, x)
        dec = np.arcsin(z)

        frame = current_geocent_frame(time)

        if scalar_input:
            ra = np.squeeze(ra)
            dec = np.squeeze(dec)
        return SkyCoord(ra, dec, frame=frame, unit='radian')

    @staticmethod
    def time_by_altitude(alt: Angle,
                         time_guess: Time,
                         location: EarthLocation,
                         timestep: float = 0.002) -> Time:
        """
        Time at which the sun crosses a particular altitude alt, which is an Angle,
        for an EarthLocation location.

        This of course happens twice a day (or not at all);
        time_guess is a Time approximating the answer.
        The usual use case will be to compute roughly when sunset or twilight occurs,
        and hand the result to this routine to get a more exact answer.

        This uses the low-precision sun location, which is typically good to 0.01 degree.
        That's plenty good enough for computing rise, set, and twilight times.

        Parameters:

        alt : Angle, single or array. If array, then must be the same length as time_guess
        Desired altitude.
        time_guess : Time, single or array
        Starting time for iteration.  This must be fairly
        close so that the iteration coverges on the correct
        phenomenon (e.g., rise time, not set time).
        location : EarthLocation

        Returns: Time if convergent
                None if non-convergent
        """
        time_guess = Time(np.asarray(time_guess.jd), format='jd')
        alt = Angle(np.asarray(alt.to_value(u.rad)), unit=u.rad)
        scalar_input = False
        if time_guess.ndim == 0 and alt.ndim == 0:
            scalar_input = True
        if time_guess.ndim == 0:
            time_guess = time_guess[None]  # Makes 1D
        if alt.ndim == 0:
            alt = alt[None]

        if len(time_guess) == 1 and len(alt) > 1:
            time_guess = Time(time_guess.jd * np.ones(len(alt)), format='jd')
        elif len(time_guess) > 1 and len(alt) == 1:
            alt = alt * np.ones(len(time_guess))
        elif len(time_guess) != len(alt):
            raise ValueError('Error: alt and time_guess have incompatible lengths')

        sun_pos = Sun.at(time_guess)
        tolerance = Angle(1.0e-4, unit=u.rad)

        delta = TimeDelta(timestep, format='jd')
   
        ha = local_sidereal_time(time_guess, location) - sun_pos.ra
        alt2, az, parang = Altitude.above(sun_pos.dec, Angle(ha, unit=u.hourangle), location.lat)

        time_guess = time_guess + delta
        sun_pos = Sun.at(time_guess)

        alt3, az, parang = Altitude.above(sun_pos.dec,
                                          local_sidereal_time(time_guess, location) - sun_pos.ra,
                                          location.lat)
        err = alt3 - alt
        deriv = (alt3 - alt2) / delta

        kount = np.zeros(len(time_guess), dtype=int)
        kk = np.where(np.logical_and(abs(err) > tolerance, kount < 10))[0][:]
        while len(kk) != 0:
            time_guess[kk] = time_guess[kk] - err[kk] / deriv[kk]
            sun_pos = Sun.at(time_guess[kk])
            alt3[kk], az[kk], parang[kk] = Altitude.above(sun_pos.dec,
                                                          local_sidereal_time(time_guess[kk], location) - sun_pos.ra,
                                                          location.lat)
            err[kk] = alt3[kk] - alt[kk]
            kount[kk] = kount[kk] + 1
            ii = np.where(kount >= 9)[0][:]
            if len(ii) != 0:
                raise ArithmeticError("Sunrise, set, or twilight calculation not converging.")
            kk = np.where(np.logical_and(abs(err) > tolerance, kount < 10))[0][:]

        if scalar_input:
            time_guess = np.squeeze(time_guess)
        return Time(time_guess, format='iso')

    @staticmethod
    def rise_and_set(location: EarthLocation,
                     time: Time,
                     midnight: Time,
                     set_alt: Angle,
                     rise_alt) -> Tuple[Time, Time, Time, Time]:
        """
        Compute rise and set times for this Sun, for the current
        ``location`` and ``time`` of the night.

        Returns
        -------
        `~astropy.time.Time`
            The time of the event for the body in the
            ``precision`` of this `Sun`.
        """

        sun_at_midnight = Sun.at(midnight)
        lst_midnight = local_sidereal_time(midnight, location)
        nt = len(time)

        twelve_twilight_alt = Angle(-12. * np.ones(nt), unit=u.deg)  # 12 degree nautical twilight

        # corresponding hr angles
        sunset_ha = hour_angle_to_angle(sun_at_midnight.dec, location.lat, set_alt)
        sunrise_ha = Angle(2. * np.pi, unit=u.rad) - hour_angle_to_angle(sun_at_midnight.dec, location.lat, rise_alt)
        twelve_twilight_ha = hour_angle_to_angle(sun_at_midnight.dec, location.lat, twelve_twilight_alt)
        sun_at_midnight_ha = (lst_midnight - sun_at_midnight.ra).wrap_at(24. * u.hour)

        sunset_guess = sun_at_midnight_ha - sunset_ha  # angles away from midnight
        sunrise_guess = sunrise_ha - sun_at_midnight_ha
        even_12twi_guess = sun_at_midnight_ha - twelve_twilight_ha
        morn_12twi_guess = Angle(2. * np.pi, unit=u.rad) - twelve_twilight_ha - sun_at_midnight_ha

        # convert to time deltas
        timedelta_sunset = TimeDelta(sunset_guess.hour / 24., format='jd')
        timedelta_sunrise = TimeDelta(sunrise_guess.hour / 24., format='jd')
        timedelta_even_12twi = TimeDelta(even_12twi_guess.hour / 24., format='jd')
        timedelta_morn_12twi = TimeDelta(morn_12twi_guess.hour / 24., format='jd')
        
        # form into times and iterate to accurate answer.
        times_sunset = midnight - timedelta_sunset  # first approx
        times_sunset = Sun.time_by_altitude(set_alt, times_sunset, location)

        times_sunrise = midnight + timedelta_sunrise  # first approx
        times_sunrise = Sun.time_by_altitude(rise_alt, times_sunrise, location)

        times_even_12twi = midnight - timedelta_even_12twi
        times_even_12twi = Sun.time_by_altitude(twelve_twilight_alt, times_even_12twi, location)

        times_morn_12twi = midnight + timedelta_morn_12twi
        times_morn_12twi = Sun.time_by_altitude(twelve_twilight_alt, times_morn_12twi, location)
        
        return times_sunrise, times_sunset, times_even_12twi, times_morn_12twi



from dataclasses import dataclass
from datetime import datetime, timedelta
from enum import auto, Enum, IntEnum
from typing import ClassVar, FrozenSet, List, Optional

from .group import AndGroup, Group, GroupID
from .observation import Observation
from .semester import Semester
from .timeallocation import TimeAllocation
from .too import TooType

ProgramID = str


class Band(IntEnum):
    """
    Program band.
    """
    BAND1 = 1
    BAND2 = 2
    BAND3 = 3
    BAND4 = 4


class ProgramMode(IntEnum):
    """
    Main operational mode, which is one of:
    * Queue
    * Classical
    * Priority Visitor (hybrid mode between queue and classical)
    """
    QUEUE = auto()
    CLASSICAL = auto()
    PV = auto()


@dataclass(frozen=True)
class ProgramType:
    """
    Represents the information encompassing the type of program.
    * abbreviation: the code used by the program type (e.g. Q, C, FT, LP)
    * name: user readable representation of the program type
    * is_science: indicates if this program type is a science program

    NOTE that ProgramType instances should NEVER be explicitly created.
    All of the valid ProgramType instances are contained in the ProgramTypes enum
    and should be accessed from there.
    """
    abbreviation: str
    name: str
    is_science: bool = True


class ProgramTypes(Enum):
    """
    A complete list of the ProgramType instances used by Gemini.
    As mentioned in ProgramType, ProgramType should never be instantiated
    outside of this enum: instead, ProgramType instances should be retrieved
    from here.
    """
    C = ProgramType('C', 'Classical')
    CAL = ProgramType('CAL', 'Calibration', False)
    DD = ProgramType('DD', "Director's Time")
    DS = ProgramType('DS', 'Demo Science')
    ENG = ProgramType('ENG', 'Engineering', False)
    FT = ProgramType('FT', 'Fast Turnaround')
    LP = ProgramType('LP', 'Large Program')
    Q = ProgramType('Q', 'Queue')
    SV = ProgramType('SV', 'System Verification')


@dataclass(unsafe_hash=True)
class Program:
    """
    Representation of a program.

    The FUZZY_BOUNDARY is a constant that allows for a fuzzy boundary for a program's
    start and end times.
    """
    id: ProgramID
    internal_id: str
    # Some programs do not have a typical name and thus cannot be associated with a semester.
    semester: Optional[Semester]
    band: Band
    thesis: bool
    mode: ProgramMode
    type: Optional[ProgramTypes]
    start: datetime
    end: datetime
    allocated_time: FrozenSet[TimeAllocation]
    root_group: AndGroup
    too_type: Optional[TooType] = None

    FUZZY_BOUNDARY: ClassVar[timedelta] = timedelta(days=14)

    def program_awarded(self) -> timedelta:
        return sum((t.program_awarded for t in self.allocated_time), timedelta())

    def program_used(self) -> timedelta:
        return sum((t.program_used for t in self.allocated_time), timedelta())

    def partner_awarded(self) -> timedelta:
        return sum((t.partner_awarded for t in self.allocated_time), timedelta())

    def partner_used(self) -> timedelta:
        return sum((t.partner_used for t in self.allocated_time), timedelta())

    def total_awarded(self) -> timedelta:
        return sum((t.total_awarded() for t in self.allocated_time), timedelta())

    def total_used(self) -> timedelta:
        return sum((t.total_used() for t in self.allocated_time), timedelta())

    def observations(self) -> List[Observation]:
        return self.root_group.observations()

    def get_group_ids(self) -> FrozenSet[GroupID]:
        return self.root_group.subgroup_ids()

    def get_group(self, group_id: GroupID) -> Optional[Group]:
        def aux(group: Group) -> Optional[Group]:
            if group.id == group_id:
                return group
            elif group.is_scheduling_group():
                for subgroup in group.children:
                    retval = aux(subgroup)
                    if retval is not None:
                        return retval
            return None

        return aux(self.root_group)



from .atom import *
from .constraints import *
from .group import *
from .magnitude import *
from .observation import *
from .observationmode import *
from .program import *
from .qastate import *
from .resource import *
from .semester import *
from .site import *
from .target import *
from .timeallocation import *
from .timingwindow import *
from .too import *

# Type alias for night indices.
NightIndex = int



from abc import abstractmethod, ABC
from dataclasses import dataclass
from datetime import timedelta
from enum import auto, Enum
from typing import FrozenSet, List, Optional, Union

from .constraints import Constraints
from .observation import Observation
from .resource import Resource
from .site import Site

from ..helpers import flatten

GroupID = str


@dataclass
class Group(ABC):
    """
    This is the base implementation of AND / OR Groups.
    Python does not allow classes to self-reference unless in static contexts,
    so we make a very simple base class to self-reference from subclasses since
    we need this functionality to allow for group nesting.

    * id: the identification of the group
    * group_name: a human-readable name of the group
    * number_to_observe: the number of children in the group that must be observed for the
      group to be considered complete
    * delay_min: used in cadences
    * delay_max: used in cadences
    """
    id: GroupID
    group_name: str
    number_to_observe: int
    delay_min: timedelta
    delay_max: timedelta
    children: Union[List['Group'], Observation]

    def __post_init__(self):
        if self.number_to_observe <= 0:
            msg = f'Group {self.group_name} specifies non-positive {self.number_to_observe} children to be observed.'
            raise ValueError(msg)

    def subgroup_ids(self) -> FrozenSet[GroupID]:
        if isinstance(self.children, Observation):
            return frozenset()
        else:
            return frozenset(subgroup.id for subgroup in self.children)

    def sites(self) -> FrozenSet[Site]:
        if isinstance(self.children, Observation):
            return frozenset([self.children.site])
        else:
            return frozenset.union(*[s.sites() for s in self.children])

    def required_resources(self) -> FrozenSet[Resource]:
        return frozenset(r for c in self.children for r in c.required_resources())

    def wavelengths(self) -> FrozenSet[float]:
        return frozenset(w for c in self.children for w in c.wavelengths())

    def constraints(self) -> FrozenSet[Constraints]:
        return frozenset(cs for c in self.children for cs in c.constraints())

    def observations(self) -> List[Observation]:
        if isinstance(self.children, Observation):
            return [self.children]
        else:
            return [o for g in self.children for o in g.observations()]

    def is_observation_group(self) -> bool:
        return issubclass(type(self.children), Observation)

    def is_scheduling_group(self) -> bool:
        return not (self.is_observation_group())

    @abstractmethod
    def is_and_group(self) -> bool:
        ...

    @abstractmethod
    def is_or_group(self) -> bool:
        ...

    def __len__(self):
        return 1 if self.is_observation_group() else len(self.children)


class AndOption(Enum):
    """
    Different options available for ordering AND group children.
    CUSTOM is used for cadences.
    """
    CONSEC_ORDERED = auto()
    CONSEC_ANYORDER = auto()
    NIGHT_ORDERED = auto()
    NIGHT_ANYORDER = auto()
    ANYORDER = auto()
    CUSTOM = auto()


@dataclass
class AndGroup(Group):
    """
    The concrete implementation of an AND group.
    It requires an AndOption to specify how its observations should be handled,
    and a previous (which should be an index into the group's children to indicate
    the previously observed child, or None if none of the children have yet been
    observed).
    """
    group_option: AndOption
    previous: Optional[int] = None

    def __post_init__(self):
        super().__post_init__()
        if self.number_to_observe != len(self.children):
            msg = f'AND group {self.group_name} specifies {self.number_to_observe} children to be observed but has ' \
                  f'{len(self.children)} children.'
            raise ValueError(msg)
        if self.previous is not None and (self.previous < 0 or self.previous >= len(self.children)):
            msg = f'AND group {self.group_name} has {len(self.children)} children and an illegal previous value of ' \
                  f'{self.previous}'
            raise ValueError(msg)

    def is_and_group(self) -> bool:
        return True

    def is_or_group(self) -> bool:
        return False

    def exec_time(self) -> timedelta:
        """
        Total execution time across the childrenn of this group.
        """
        if issubclass(type(self.children), Observation):
            return self.children.exec_time()
        else:
            sum(child.exec_time() for child in self.children)

    def total_used(self) -> timedelta:
        """
        Total time used across the group: includes program time and partner time.
        """
        if issubclass(type(self.children), Observation):
            return self.children.total_used()
        else:
            sum(child.total_used() for child in self.children)

    def instruments(self) -> FrozenSet[Resource]:
        """
        Return a set of all instruments used in this group.
        """
        if issubclass(type(self.children), Observation):
            instrument = self.children.instrument()
            if instrument is not None:
                return frozenset({instrument})
            else:
                return frozenset()
        else:
            return frozenset(flatten([child.instruments() for child in self.children]))


@dataclass
class OrGroup(Group):
    """
    The concrete implementation of an OR group.
    The restrictions on an OR group is that it must explicitly require not all
    of its children to be observed.
    """

    def __post_init__(self):
        super().__post_init__()
        if self.number_to_observe >= len(self.children):
            msg = f'OR group {self.group_name} specifies {self.number_to_observe} children to be observed but has ' \
                  f'{len(self.children)} children.'
            raise ValueError(msg)

    def is_and_group(self) -> bool:
        return False

    def is_or_group(self) -> bool:
        return True



# This has to be outside of Observation to avoid circular imports.
from enum import Enum


class ObservationMode(str, Enum):
    """
    TODO: This is not stored anywhere and is only used temporarily in the atom code in the
    TODO: OcsProgramExtractor. Should it be stored anywhere or is it only used in intermediate
    TODO: calculations? It seems to depend on the instrument and FPU.
    """
    UNKNOWN = 'unknown'
    IMAGING = 'imaging'
    LONGSLIT = 'longslit'
    IFU = 'ifu'
    MOS = 'mos'
    XD = 'xd'
    CORON = 'coron'
    NRM = 'nrm'



from abc import ABC
from datetime import timedelta
from typing import FrozenSet, NoReturn, Optional

from astropy.time import Time

# TODO: No type information here because of circular dependencies between common.minimodel and this class.
# TODO: Introduces circular dependencies
# from common.minimodel import Resource, ObservationMode

# TODO: module common has no attribute minimodel
# import common.minimodel


class ObservatoryProperties(ABC):
    """
    Observatory-specific methods that are not tied to other components or
    structures, and allow computations to be implemented in one place.
    """
    _properties: Optional['ObservatoryProperties'] = None

    @staticmethod
    def set_properties(cls) -> NoReturn:
        if not issubclass(cls, ObservatoryProperties):
            raise ValueError('Illegal properties value.')
        ObservatoryProperties._properties = cls()

    @staticmethod
    def _check_properties() -> NoReturn:
        if ObservatoryProperties._properties is None:
            raise ValueError('Properties have not been set.')

    @staticmethod
    def determine_standard_time(resources: FrozenSet,
                                wavelengths: FrozenSet[float],
                                modes: FrozenSet,
                                cal_length: int) -> Time:
        """
        Given the information, determine the length in hours required for calibration
        on a standard star.
        TODO: Is this the correct description?
        TODO: Do we need modes here or can these just be determined from resources?
        TODO: Based on the Gemini code, it seems like the latter is the case, but we do have
        TODO: the obsmode code in the atom extraction which provides an ObservationMode.
        """
        ObservatoryProperties._check_properties()
        return ObservatoryProperties._properties.determine_standard_time(
            resources,
            wavelengths,
            modes,
            cal_length
        )

    @staticmethod
    def is_instrument(resource) -> bool:
        """
        Determine if the given resource is an instrument or not.
        """
        ObservatoryProperties._check_properties()
        return ObservatoryProperties._properties.is_instrument(resource)

    @staticmethod
    def acquisition_time(resource, observation_mode) -> Optional[timedelta]:
        """
        Given a resource, check if it is an instrument, and if so, lookup the
        acquisition time for the specified mode.
        """
        ObservatoryProperties._check_properties()
        return ObservatoryProperties._properties.acquisition_time(resource, observation_mode)



from dataclasses import dataclass
from datetime import timedelta
from typing import FrozenSet

from .qastate import QAState
from .resource import Resource


@dataclass
class Atom:
    """
    Atom information, where an atom is the smallest schedulable set of steps
    such that useful science can be obtained from performing them.
    Wavelengths must be specified in microns.
    """
    id: int
    exec_time: timedelta
    prog_time: timedelta
    part_time: timedelta
    observed: bool
    qa_state: QAState
    guide_state: bool
    resources: FrozenSet[Resource]
    wavelengths: FrozenSet[float]



from enum import Enum
from typing import Optional

import astropy.coordinates
import pytz


class Site(Enum):
    """
    The sites belonging to the observatory using the Scheduler.

    This will have to be customized by a given observatory if used independently
    of Gemini.
    """
    GN = ('Gemini North', '568@399')
    GS = ('Gemini South', 'I11@399')

    def __init__(self, site_name: str, coordinate_center: str, astropy_lookup: Optional[str] = None):
        self.site_name = site_name
        self.coordinate_center = coordinate_center

        if astropy_lookup is None:
            astropy_lookup = site_name.lower().replace(' ', '_')

        try:
            self.location = astropy.coordinates.EarthLocation.of_site(astropy_lookup)
        except astropy.coordinates.UnknownSiteException as e:
            msg = f'Unknown site lookup: {astropy_lookup}.'
            raise ValueError(e, msg)

        timezone_info = self.location.info.meta['timezone']
        try:
            self.timezone = pytz.timezone(timezone_info)
        except pytz.UnknownTimeZoneError as e:
            msg = f'Unknown time zone lookup: {timezone_info}.'
            raise ValueError(e, msg)


ALL_SITES = frozenset(s for s in Site)



import numpy as np


def scalar_input(func):
    """
    Decorator to convert a function that returns a tuple to a function that returns a scalar.
    TODO: TO IMPLEMENT THIS IS NECESSARY TO REMOVE ASTROPY outputs or made them numpy compatible.
    """

    def wrapper(*args, **kwargs):

        # transform the input to numpy
        np_args = [np.asarray(arg) for arg in args]
        if any(arg.ndim == 0 for arg in np_args):
            args = [arg[np.newaxis] for arg in np_args]
           
        return np.squeeze(func(*args, **kwargs))
    return wrapper



from enum import auto, IntEnum


class QAState(IntEnum):
    """
    These correspond to the QA States in the OCS for Observations.
    Entries in the obs log should be made uppercase for lookups into
    this enum.
    """
    NONE = auto()
    UNDEFINED = auto()
    FAIL = auto()
    USABLE = auto()
    PASS = auto()
    # TODO: Not in original mini-model description, but returned by OCS.
    CHECK = auto()



"""
All code in this package is a refactored, numpy-vectorized version of thorskyutil.py:

https://github.com/jrthorstensen/thorsky/blob/master/thorskyutil.py

utility and miscellaneous time and the sky routines built mostly on astropy.

Copyright John Thorstensen, 2018; offered under the GNU Public License 3.0.

Vectorized by Bryan Miller, Gemini Observatory.
Refactored and clarified by Sergio Troncoso, Gemini Observatory.
Modified by Sebastian Raaphorst, Gemini Observatory, to remove all deviations from Python 3.x PEP8 style guide.
"""

from .altitude import *
from .brightness import *
from .constants import *
from .events import *
from .moon import *
from .sun import *
from .utils import *



from abc import ABC
from dataclasses import dataclass
from enum import auto, Enum, IntEnum
from typing import Set

import numpy.typing as npt

from .magnitude import Magnitude

TargetName = str


class TargetType(Enum):
    """
    The type associated with a target in an observation.
    """
    BASE = auto()
    USER = auto()
    BLIND_OFFSET = auto()
    OFF_AXIS = auto()
    TUNING_STAR = auto()
    GUIDESTAR = auto()
    OTHER = auto()


class GuideSpeed(IntEnum):
    """
    How quickly a guider can guide on a guide star.
    """
    SLOW = auto()
    MEDIUM = auto()
    FAST = auto()


class TargetTag(Enum):
    """
    A tag used by nonsidereal targets to indicate their type.
    """
    COMET = auto()
    ASTEROID = auto()
    MAJOR_BODY = auto()


@dataclass
class Target(ABC):
    """
    Basic target information.
    """
    name: TargetName
    magnitudes: Set[Magnitude]
    type: TargetType

    def guide_speed(self) -> GuideSpeed:
        """
        Calculate the guide speed for this target.
        """
        ...


@dataclass
class SiderealTarget(Target):
    """
    For a SiderealTarget, we have an RA and Dec and then proper motion information
    to calculate the exact position.

    RA and Dec should be specified in decimal degrees.
    Proper motion must be specified in milliarcseconds / year.
    Epoch must be the decimal year.

    NOTE: The proper motion adjusted coordinates can be found in the TargetInfo in coord.
    """
    ra: float
    dec: float
    pm_ra: float
    pm_dec: float
    epoch: float


@dataclass
class NonsiderealTarget(Target):
    """
    For a NonsiderealTarget, we have a HORIZONS designation to indicate the lookup
    information, a tag to determine the type of target, and arrays of ephemerides
    to specify the position.
    """
    des: str
    tag: TargetTag
    ra: npt.NDArray[float]
    dec: npt.NDArray[float]



from dataclasses import dataclass
from enum import auto, Enum
from typing import Optional


class MagnitudeSystem(Enum):
    """
    List of magnitude systems associated with magnitude bands.
    """
    VEGA = auto()
    AB = auto()
    JY = auto()


@dataclass(frozen=True)
class MagnitudeBand:
    """
    THIS CLASS SHOULD NOT BE INSTANTIATED.
    They are fully enumerated in MagnitudeBands, so they should be looked up by name there.

    Values for center and width are specified in microns.
    """
    name: str
    center: float
    width: float
    system: MagnitudeSystem = MagnitudeSystem.VEGA
    description: Optional[str] = None


class MagnitudeBands(Enum):
    """
    It is unconventional to use lowercase characters in an enum, but to differentiate
    them from the uppercase magnitude bands, we must.

    Look up the MagnitudeBand from this Enum as follows:
    MagnitudeBands[name]
    """
    u = MagnitudeBand('u', 0.356, 0.046, MagnitudeSystem.AB, 'UV')
    g = MagnitudeBand('g', 0.483, 0.099, MagnitudeSystem.AB, 'green')
    r = MagnitudeBand('r', 0.626, 0.096, MagnitudeSystem.AB, 'red')
    i = MagnitudeBand('i', 0.767, 0.106, MagnitudeSystem.AB, 'far red')
    z = MagnitudeBand('z', 0.910, 0.125, MagnitudeSystem.AB, 'near-infrared')
    U = MagnitudeBand('U', 0.360, 0.075, description='ultraviolet')
    B = MagnitudeBand('B', 0.440, 0.090, description='blue')
    V = MagnitudeBand('V', 0.550, 0.085, description='visual')
    UC = MagnitudeBand('UC', 0.610, 0.063, description='UCAC')
    R = MagnitudeBand('R', 0.670, 0.100, description='red')
    I = MagnitudeBand('I', 0.870, 0.100, description='infrared')
    Y = MagnitudeBand('Y', 1.020, 0.120)
    J = MagnitudeBand('J', 1.250, 0.240)
    H = MagnitudeBand('H', 1.650, 0.300)
    K = MagnitudeBand('K', 2.200, 0.410)
    L = MagnitudeBand('L', 3.760, 0.700)
    M = MagnitudeBand('M', 4.770, 0.240)
    N = MagnitudeBand('N', 10.470, 5.230)
    Q = MagnitudeBand('Q', 20.130, 1.650)
    AP = MagnitudeBand('AP', 0.550, 0.085, description='apparent')


@dataclass(frozen=True)
class Magnitude:
    """
    A magnitude value in a particular band.
    """
    band: MagnitudeBands
    value: float
    error: Optional[float] = None



# Basic type aliases for usefulness.
from typing import List, TypeVar, Union
import numpy.typing as npt
from astropy.time import Time

T = TypeVar('T')


ScalarOrNDArray = Union[T, npt.NDArray[T]]
TimeScalarOrNDArray = Union[Time, npt.NDArray[float]]
ListOrNDArray = Union[List[T], npt.NDArray[T]]



from dataclasses import dataclass
from typing import Optional


@dataclass(frozen=True)
class Resource:
    """
    This is a general observatory resource.
    It can consist of a guider, an instrument, or a part of an instrument,
    or even a personnel and is used to determine what observations can be
    performed at a given time based on the resource availability.
    """
    id: str
    description: Optional[str] = None

    def __post_init__(self):
        if self.id is None or 'NONE' in self.id.upper():
            raise ValueError('Should not have any Resources equal to None or containing "None"')

    def __eq__(self, other):
        return isinstance(other, Resource) and self.id == other.id

    def __repr__(self):
        return f'Resource(id=\'{self.id}\')'




from ..minimodel import Group, Observation, Program
from typing import NoReturn


def print_program(program: Program) -> NoReturn:
    """
    Print the high-level information about a program in human semi-readable format
    to give an idea as to its structure.
    """
    print(f'Program: {program.id}')

    def sep(depth: int) -> str:
        return '----- ' * depth

    def print_observation(depth: int, obs: Observation) -> NoReturn:
        print(f'{sep(depth)} Observation: {obs.id}')
        for atom in obs.sequence:
            print(f'{sep(depth + 1)} {atom}')

    def print_group(depth: int, group: Group) -> NoReturn:
        # Is this a subgroup or an observation?
        if isinstance(group.children, Observation):
            print_observation(depth, group.children)
        elif isinstance(group.children, list):
            print(f'{sep(depth)} Group: {group.id}')
            for child in group.children:
                print_group(depth + 1, child)

    # Print the group and atom information.
    print_group(1, program.root_group)



from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import ClassVar, Optional


@dataclass(frozen=True)
class TimingWindow:
    """
    Representation of timing windows in the mini-model.

    For infinite duration, set duration to timedelta.max.
    For repeat, -1 means forever repeating, 0 means non-repeating.
    For period, None should be used if repeat < 1.
    """
    start: datetime
    duration: timedelta
    repeat: int
    period: Optional[timedelta]

    # For infinite duration, use the length of an LP.
    INFINITE_DURATION_FLAG: ClassVar[int] = -1
    INFINITE_DURATION: ClassVar[int] = timedelta(days=3 * 365, hours=24)
    FOREVER_REPEATING: ClassVar[int] = -1
    NON_REPEATING: ClassVar[int] = 0
    NO_PERIOD: ClassVar[Optional[timedelta]] = None

    # A number to be used by the Scheduler to represent infinite repeats from the
    # perspective of the OCS: if FOREVER_REPEATING is selected, then it is converted
    # into this for calculation purposes.
    OCS_INFINITE_REPEATS: ClassVar[int] = 1000



from dataclasses import dataclass
from datetime import timedelta
from enum import auto, IntEnum
from typing import List, Mapping, Optional, Set

from .atom import Atom
from .constraints import Constraints
from .qastate import QAState
from .resource import Resource
from .site import Site
from .target import Target, TargetType
from .too import TooType

ObservationID = str


class ObservationStatus(IntEnum):
    """
    The status of an observation as indicated in the Observing Tool / ODB.
    """
    NEW = auto()
    INCLUDED = auto()
    PROPOSED = auto()
    APPROVED = auto()
    FOR_REVIEW = auto()
    # TODO: Not in original mini-model description, but returned by OCS.
    ON_HOLD = auto()
    READY = auto()
    ONGOING = auto()
    OBSERVED = auto()
    INACTIVE = auto()
    PHASE2 = auto()


class Priority(IntEnum):
    """
    An observation's priority.
    Note that these are ordered specifically so that we can compare them.
    """
    LOW = auto()
    MEDIUM = auto()
    HIGH = auto()


class SetupTimeType(IntEnum):
    """
    The setup time type for an observation.
    """
    NONE = auto()
    REACQUISITION = auto()
    FULL = auto()


class ObservationClass(IntEnum):
    """
    The class of an observation.

    Note that the order of these is specific and deliberate: they are listed in
    preference order for observation classes, and hence, should not be rearranged.
    These correspond to the values in the OCS when made uppercase.
    """
    SCIENCE = auto()
    PROGCAL = auto()
    PARTNERCAL = auto()
    ACQ = auto()
    ACQCAL = auto()
    DAYCAL = auto()


@dataclass
class Observation:
    """
    Representation of an observation.
    Non-obvious fields are documented below.
    * id should represent the observation's ID, e.g. GN-2018B-Q-101-123.
    * internal_id is the key associated with the observation
    * order refers to the order of the observation in either its group or the program
    * targets should contain a complete list of all targets associated with the observation,
      with the base being in the first position
    * guiding is a map between guide probe resources and their targets
    """
    id: ObservationID
    internal_id: str
    order: int
    title: str
    site: Site
    status: ObservationStatus
    active: bool
    priority: Priority
    setuptime_type: SetupTimeType
    acq_overhead: timedelta

    # TODO: This will be handled differently between OCS and GPP.
    # TODO: 1. In OCS, when the sequence is examined, the ObservationClasses of the
    # TODO:    individual observes (sequence steps) will be analyzed and the highest
    # TODO:    precedence one will be set for the observation (based on the earliest
    # TODO:    ObservationClass in the enum).
    # TODO: 2. In GPP, this information will be handled automatically and require no
    # TODO:    special processing.
    # TODO: Should this be Optional?
    obs_class: ObservationClass

    targets: List[Target]
    guiding: Mapping[Resource, Target]
    sequence: List[Atom]

    # Some observations do not have constraints, e.g. GN-208A-FT-103-6.
    constraints: Optional[Constraints]

    too_type: Optional[TooType] = None

    def base_target(self) -> Optional[Target]:
        """
        Get the base target for this Observation if it has one, and None otherwise.
        """
        return next(filter(lambda t: t.type == TargetType.BASE, self.targets), None)

    def exec_time(self) -> timedelta:
        """
        Total execution time for the program, which is the sum across atoms and the acquisition overhead.
        """
        return sum((atom.exec_time for atom in self.sequence), timedelta()) + self.acq_overhead

    def total_used(self) -> timedelta:
        """
        Total program time used: includes program time and partner time.
        """
        return self.program_used() + self.partner_used()

    def required_resources(self) -> Set[Resource]:
        """
        The required resources for an observation based on the sequence's needs.
        """
        return self.guiding.keys() | {r for a in self.sequence for r in a.resources}

    def instrument(self) -> Optional[Resource]:
        """
        Returns a resource that is an instrument, if one exists.
        There should be only one.
        """
        return next(filter(lambda r: api.observatory.abstract.ObservatoryProperties.is_instrument(r),
                           self.required_resources()), None)

    def wavelengths(self) -> Set[float]:
        """
        The set of wavelengths included in the sequence.
        """
        return {w for c in self.sequence for w in c.wavelengths}

    def constraints(self) -> Set[Constraints]:
        """
        A set of the constraints required by the observation.
        In the case of an observation, this is just the (optional) constraints.
        """
        return {self.constraints} if self.constraints is not None else {}

    def program_used(self) -> timedelta:
        """
        We roll this information up from the atoms as it will be calculated
        during the GreedyMax algorithm. Note that it is also available directly
        from the OCS, which is used to populate the time allocation.
        """
        return sum((atom.prog_time for atom in self.sequence), start=timedelta())

    def partner_used(self) -> timedelta:
        """
        We roll this information up from the atoms as it will be calculated
        during the GreedyMax algorithm. Note that it is also available directly
        from the OCS, which is used to populate the time allocation.
        """
        return sum((atom.part_time for atom in self.sequence), start=timedelta())

    @staticmethod
    def _select_obsclass(classes: List[ObservationClass]) -> Optional[ObservationClass]:
        """
        Given a list of non-empty ObservationClasses, determine which occurs with
        highest precedence in the ObservationClasses enum, i.e. has the lowest index.

        This will be used when examining the sequence for atoms.

        TODO: Move this to the ODB program extractor as the logic is used there.
        TODO: Remove from Bryan's atomizer.
        """
        return min(classes, default=None)

    @staticmethod
    def _select_qastate(qastates: List[QAState]) -> Optional[QAState]:
        """
        Given a list of non-empty QAStates, determine which occurs with
        highest precedence in the QAStates enum, i.e. has the lowest index.

        TODO: Move this to the ODB program extractor as the logic is used there.
        TODO: Remove from Bryan's atomizer.
        """
        return min(qastates, default=None)

    def __len__(self):
        """
        This is to treat observations the same as groups and is a bit of a hack.
        Observations are to be placed in AND Groups of size 1 for scheduling purposes.
        """
        return 1

    def __eq__(self, other: 'Observation') -> bool:
        """
        We override the equality checker created by @dataclass to temporarily skip sequence
        comparison in test cases until the atom creation process is finish.
        """

        return (dict((k, v) for k, v in self.__dict__.items() if k != 'sequence') ==
                dict((k, v) for k, v in other.__dict__.items() if k != 'sequence'))



from typing import Final

import astropy.units as u
from astropy.coordinates.distances import Distance
from astropy.time import Time

"""
Constants that are used here and there.  Some are Quantities,
others are just floats. Not all are used.

The planet-coefs are for series expansions for the phase functions
of the planets, used in predicting apparent magnitude. See code.
"""

PI: Final[float] = 3.14159265358979
TWOPI: Final[float] = 6.28318530717959
PI_OVER_2: Final[float] = 1.57079632679490
ARCSEC_IN_RADIAN: Final[float] = 206264.8062471
DEG_IN_RADIAN: Final[float] = 57.2957795130823
HRS_IN_RADIAN: Final[float] = 3.819718634205
KMS_AUDAY: Final[float] = 1731.45683633  # /* km per sec in 1 AU/day */
SPEED_OF_LIGHT: Final[float] = 299792.458  # /* in km per sec ... exact. */
SS_MASS: Final[float] = 1.00134198  # /* solar system mass in solar units */
J2000: Final[float] = 2451545.  # /* Julian date at standard epoch */
J2000_Time: Final[Time] = Time(2451545., format='jd')  # J2000 rendered as a Time
SEC_IN_DAY: Final[float] = 86400.
FLATTEN: Final[float] = 0.003352813  # /* flattening of earth, 1/298.257 */
EQUAT_RAD: Final[Distance] = 6378137. * u.m  # /* equatorial radius of earth, meters */
EARTHRAD_IN_AU: Final[float] = 23454.7910556298  # /* number of earth rad in 1 au */
ASTRO_UNIT: Final[float] = 1.4959787066e11  # /* 1 AU in meters */
RSUN: Final[float] = 6.96000e8  # /* IAU 1976 recom. solar radius, meters */
RMOON: Final[float] = 1.738e6  # /* IAU 1976 recom. lunar radius, meters */
PLANET_TOL: Final[float] = 3.  # /* flag if nearer than 3 degrees
KZEN: Final[float] = 0.172  # mag / airmass relation for Hale Pohaku



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lucupy',
 'lucupy.decorators',
 'lucupy.helpers',
 'lucupy.minimodel',
 'lucupy.observatory.abstract',
 'lucupy.observatory.gemini',
 'lucupy.output',
 'lucupy.sky',
 'lucupy.timeutils',
 'lucupy.types']

package_data = \
{'': ['*']}

install_requires = \
['astropy>=5.1,<6.0', 'pytz>=2022.2,<2023.0']

setup_kwargs = {
    'name': 'lucupy',
    'version': '0.1.1',
    'description': 'A striped down version of the lucuma core in Python for Schedule',
    'long_description': '',
    'author': 'Sergio Troncoso',
    'author_email': 'sergio.troncoso@noirlab.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/stroncod/lucupy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from datetime import timedelta
from enum import Enum, EnumMeta
from typing import FrozenSet, Optional

import astropy.units as u
from astropy.time import Time

from ..abstract import ObservatoryProperties
from ...minimodel import ObservationMode
from ...minimodel import Resource


class GeminiProperties(ObservatoryProperties):
    """
    Implementation of ObservatoryCalculations specific to Gemini.
    """
    class _InstrumentsMeta(EnumMeta):
        def __contains__(cls, r: Resource) -> bool:
            return any(inst.value.id in r.id for inst in cls.__members__.values())

    # Gemini-specific instruments.
    class Instruments(Enum, metaclass=_InstrumentsMeta):
        FLAMINGOS2 = Resource('Flamingos2')
        GNIRS = Resource('GNIRS')
        NIFS = Resource('NIFS')
        IGRINS = Resource('IGRINS')
        GMOS_S = Resource('GMOS-S')
        GMOS_N = Resource('GMOS-N')

    # Instruments for which there are set standards.
    _STANDARD_INSTRUMENTS = [Instruments.FLAMINGOS2,
                             Instruments.GNIRS,
                             Instruments.NIFS,
                             Instruments.IGRINS]

    @staticmethod
    def determine_standard_time(resources: FrozenSet[Resource],
                                wavelengths: FrozenSet[float],
                                modes: FrozenSet[ObservationMode],
                                cal_length: int) -> Time:
        """
        Determine the standard star time required for Gemini.
        """
        if cal_length > 1:
            # Check to see if any of the resources are instruments.
            # TODO: We may only want to include specific resources, in which case, modify
            # TODO: Instruments above to be StandardInstruments.
            if any(resource in GeminiProperties._STANDARD_INSTRUMENTS for resource in resources):
                if all(wavelength <= 2.5 for wavelength in wavelengths):
                    return 1.5 * u.h
                else:
                    return 1.0 * u.h
            if ObservationMode.IMAGING in modes:
                return 2.0 * u.h
            return 0.0 * u.h

    @staticmethod
    def is_instrument(resource: Resource) -> bool:
        return resource in GeminiProperties.Instruments

    @staticmethod
    def acquisition_time(resource: Resource, observation_mode: ObservationMode) -> Optional[timedelta]:
        if not GeminiProperties.is_instrument(resource):
            return None
        ...



from dataclasses import dataclass
from enum import Enum


class SemesterHalf(str, Enum):
    """
    Gemini typically schedules programs for two semesters per year, namely A and B.
    For other observatories, this logic might have to be substantially changed.
    """
    A = 'A'
    B = 'B'


@dataclass(frozen=True, order=True)
class Semester:
    """
    A semester is a period for which programs may be submitted to Gemini and consists of:
    * A four digit year
    * Two semesters during each year, indicated by the SemesterHalf
    """
    year: int
    half: SemesterHalf

    def __str__(self):
        return f'{self.year}{self.half.value}'
